"""
Setup file for PassTheSalt.
"""

import io
import os
import re

from setuptools import find_packages, setup


def get_metadata():
    """
    Return metadata for PassTheSalt.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    init_path = os.path.join(here, 'src', 'passthesalt', '__init__.py')
    readme_path = os.path.join(here, 'README.md')

    with io.open(init_path, encoding='utf-8') as f:
        about_text = f.read()

    metadata = {
        key: re.search(r'__' + key + r"__ = '(.*?)'", about_text).group(1)
        for key in (
            'title',
            'version',
            'url',
            'author',
            'author_email',
            'license',
            'description',
        )
    }
    metadata['name'] = metadata.pop('title')

    with io.open(readme_path, encoding='utf-8') as f:
        metadata['long_description'] = f.read()
        metadata['long_description_content_type'] = 'text/markdown'

    return metadata


metadata = get_metadata()

# Primary requirements
install_requires = [
    'click        ==7.*',
    'cryptography ==2.*',
    'pyperclip    ==1.*',
    'requests     ==2.*',
    'serde[ext]   ==0.8.*',
    'tabulate     ==0.8.*',
    'toml         ==0.10.*',
]
entry_points = {
    'console_scripts': ['pts=passthesalt.cli:cli', 'passthesalt=passthesalt.cli:cli']
}

setup(
    # Options
    install_requires=install_requires,
    python_requires='>=3.6',
    packages=find_packages('src'),
    entry_points=entry_points,
    package_dir={'': 'src'},
    py_modules=['passthesalt'],
    # Metadata
    download_url='{url}/archive/{version}.tar.gz'.format(**metadata),
    project_urls={'Issue Tracker': '{url}/issues'.format(**metadata)},
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    keywords='password manager pbkdf2',
    **metadata,
)



"""
Extensions for serde for use in PassTheSalt.
"""

import datetime
from base64 import b64decode, b64encode

import toml
from serde import Model as BaseModel
from serde import fields


class DateTime(fields.DateTime):
    """
    A custom `DateTime` `Field` that uses multiple valid formats.
    """

    formats = ('%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d')

    def __init__(self, **kwargs):
        """
        Create a new `DateTime`.

        Args:
            **kwargs: keyword arguments for the `Field` constructor.
        """
        super().__init__(**kwargs)
        del self.format

    def serialize(self, value):
        """
        Serialize the given `~datetime.datetime` as a string.

        Args:
            value (~datetime.datetime): the datetime object to serialize.

        Returns:
            str: a string representation of the datetime.
        """
        if value.microsecond:
            format = self.formats[0]
        elif value.hour or value.minute or value.second:
            format = self.formats[1]
        else:
            format = self.formats[2]

        return value.strftime(format)

    def deserialize(self, value):
        """
        Deserialize the given string as a `~datetime.datetime` using one of the
        valid formats.

        Args:
            value (str): the string to deserialize.

        Returns:
            ~datetime.datetime: the deserialized datetime.
        """
        for datetime_format in self.formats:
            try:
                return datetime.datetime.strptime(value, datetime_format)
            except ValueError:
                pass

        raise ValueError(f'datetime {value!r} does not match any valid formats')


class Model(BaseModel):
    """
    A custom Model that has a modified Field.
    """

    modified: fields.Optional(DateTime, default=datetime.datetime.utcnow)

    def to_base64(self, **kwargs):
        """
        Dump the model as a JSON string and base64 encode it.

        Args:
            **kwargs: extra keyword arguments to pass directly to `json.dumps`.

        Returns:
            str: a base64 encoded representation of this model.
        """
        return b64encode(self.to_json(**kwargs).encode()).decode('ascii')

    def to_toml(self, **kwargs):
        """
        Dump the model as a TOML string.

        Args:
            **kwargs: extra keyword arguments to pass directly to `toml.dumps`.

        Returns:
            str: a TOML representation of this model.
        """
        return toml.dumps(self.to_dict(), **kwargs)

    def to_path(self, p, dict=None, **kwargs):
        """
        Dump the model to a file path.

        Args:
            p (str): the file path to write to.
            **kwargs: extra keyword arguments to pass directly to `json.dumps`.
        """
        with open(p, 'w') as f:
            f.write(self.to_json(**kwargs))

    @classmethod
    def from_base64(cls, s, **kwargs):
        """
        Load the model from a base64 encoded string.

        Args:
            s (str): the JSON string.
            **kwargs: extra keyword arguments to pass directly to `json.loads`.

        Returns:
            Model: an instance of this model.            .
        """
        return cls.from_json(b64decode(s.encode()).decode('utf-8'), **kwargs)

    @classmethod
    def from_toml(cls, s, **kwargs):
        """
        Load the model from a TOML string.

        Args:
            s (str): the TOML string.
            **kwargs: extra keyword arguments to pass directly to `toml.loads`.

        Returns:
            Model: an instance of this model.
        """
        return cls.from_dict(toml.loads(s, **kwargs))

    @classmethod
    def from_path(cls, p, **kwargs):
        """
        Load the model from a file path.

        Args:
            p (str): the file path to read from.
            **kwargs: extra keyword arguments to pass directly to `json.loads`.

        Returns:
            Model: an instance of this model.            .
        """
        with open(p) as f:
            return cls.from_json(f.read(), **kwargs)

    def touch(self):
        """
        Update the modified time of this object to the current UTC time.
        """
        self.modified = datetime.datetime.utcnow()
